import yargs from "yargs";
import { hideBin } from "yargs/helpers";
import dotenv from "dotenv";
import { config, getConfig, prompt, schema } from "@next-dev-utils/utils/config";
import fs from "node:fs/promises";
import path from "node:path";
import stripIndent from "strip-indent";
import { existsSync, promises } from "node:fs";
import ora from "ora";
import { execa } from "execa";
import { next, node, pack, packNext, pnpm } from "@next-dev-utils/utils";
import clipboard from "clipboardy";

//#region src/commands/config.ts
async function configCommand(args) {
	if (args.operation === "get") {
		if (!args.key) {
			for (const key in schema) console.log(`${key}: ${config.get(key)}`);
			return;
		}
		console.log(config.get(args.key));
	} else if (args.value) {
		if (!args.key) throw new Error("key is required when setting a value");
		config.set(args.key, args.value);
	} else {
		if (!args.key) throw new Error("key is required when setting a value");
		const input = await prompt(args.key);
		config.set(args.key, input);
	}
}

//#endregion
//#region src/commands/create-reproduction.ts
async function createReproductionCommand(args) {
	const reproduction = path.join(await getConfig("next_project_path"), "__reproductions__", args.name);
	await fs.mkdir(reproduction, { recursive: true });
	await fs.writeFile(path.join(reproduction, "next.config.js"), stripIndent(`
    module.exports = {
        experimental: {},
    };`));
	await fs.mkdir(path.join(reproduction, "app"), { recursive: true });
	console.log("Created reproduction folder", reproduction);
}

//#endregion
//#region src/commands/debug.ts
async function removeDotNextDirectory(nextProjectPath) {
	const spinner = ora("Removing existing .next directory").start();
	try {
		await promises.rm(path.join(nextProjectPath, ".next"), {
			recursive: true,
			force: true
		});
	} catch (err) {
		spinner.fail("Failed to remove existing .next directory");
		throw err;
	}
	spinner.succeed();
}
async function debugCommand(options) {
	const nextProjectPath = options.nextProjectDirectory;
	if (!existsSync(nextProjectPath)) throw new Error(`The next project directory ${nextProjectPath} does not exist.`);
	if (options.run && options.mode !== "start" && options.mode !== "prod" && options.mode !== "standalone") throw new Error("The --run option is only valid with the start, prod, or standalone command");
	if (options.mode !== "start" && options.mode !== "export") await removeDotNextDirectory(nextProjectPath);
	const controller = new AbortController();
	const { signal } = controller;
	process.on("SIGINT", async () => {
		controller.abort();
		process.exit(0);
	});
	process.on("SIGTERM", async () => {
		controller.abort();
		process.exit(0);
	});
	const nodeOptions = [];
	try {
		if (options.mode === "build" || options.mode === "prod" || options.mode === "standalone") await next(["build", nextProjectPath], {
			stdio: "inherit",
			signal
		});
		if (options.mode === "start" || options.mode === "prod" || options.mode === "standalone" || options.mode === "dev") {
			let start;
			switch (options.mode) {
				case "start":
				case "prod":
					start = next(["start", nextProjectPath], {
						stdio: "inherit",
						signal,
						nodeOptions
					});
					break;
				case "dev":
					start = next(["dev", nextProjectPath], {
						stdio: "inherit",
						signal,
						nodeOptions
					});
					break;
				case "standalone":
					start = node([path.join(nextProjectPath, ".next", "standalone", "server.js")], {
						stdio: "inherit",
						signal
					});
					break;
			}
			let run;
			if (options.run) {
				await new Promise((resolve) => setTimeout(resolve, 1e3));
				run = execa(options.run, {
					shell: true,
					stdio: "inherit",
					signal
				});
				run.finally(() => {
					if (controller.signal.aborted) return;
					console.log("[next-dev-utils] --run finished, stopping server");
					controller.abort();
				});
				start.finally(() => {
					if (controller.signal.aborted) return;
					console.log("[next-dev-utils] server finished, stopping --run");
					controller.abort();
				});
			}
			await Promise.all([start, run]);
		} else if (options.mode === "export") await next(["export", nextProjectPath], {
			stdio: "inherit",
			signal
		});
	} catch (err) {
		if (!err.isCanceled) throw err;
	}
	if (options.rm) await removeDotNextDirectory(nextProjectPath);
}

//#endregion
//#region src/commands/pack.ts
async function packCommand(options) {
	if (options.json && options.serve) throw new Error("Cannot use --json and --serve together");
	const url = await pack({
		...options,
		progress: options.json || options.progress
	});
	await clipboard.write(url);
	if (!options.json) console.log("\nCopied URL to clipboard 🦄");
	if (options.json) console.log(url);
}

//#endregion
//#region src/commands/pack-next.ts
async function packNextCommand(options) {
	if (options.json && options.serve) throw new Error("Cannot use --json and --serve together");
	if (options.json && options.install) throw new Error("Cannot use --json and --install together");
	let controller;
	if (options.install) {
		controller = new AbortController();
		options.serve = true;
	}
	const url = await packNext({
		...options,
		progress: options.json || options.install,
		signal: controller?.signal
	});
	if (!options.json && !options.install) {
		await clipboard.write(url);
		console.log("\nCopied URL to clipboard 🦄");
	} else if (options.json) console.log(url);
	else if (options.install) {
		await pnpm([
			"add",
			"--prefer-offline",
			JSON.stringify(url)
		]);
		controller?.abort();
	}
}

//#endregion
//#region src/commands/test-deploy.ts
async function testDeployCommand(options) {
	const nextProjectPath = await getConfig("next_project_path");
	let testFile = options["test-file"];
	if (!path.isAbsolute(testFile)) testFile = path.join(process.cwd(), testFile);
	if (!existsSync(testFile)) throw new Error(`The test file ${options["test-file"]} does not exist. Please specify a valid test file.`);
	const e2eDir = path.join(nextProjectPath, "test", "e2e");
	if (!testFile.startsWith(e2eDir + path.sep)) {
		console.log(e2eDir);
		throw new Error(`The test file ${options["test-file"]} is not in a test/e2e directory. This can only be used with e2e tests.`);
	}
	if (!/\.test\.(js|ts)$/.test(testFile)) throw new Error(`The test file ${options["test-file"]} does not match the pattern *.test.{js,ts}. Please specify a valid test file.`);
	const VERCEL_TEST_TEAM = await getConfig("vercel_test_team");
	const VERCEL_TEST_TOKEN = await getConfig("vercel_test_token");
	const url = await packNext({ progress: true });
	await pnpm(["test-deploy", path.relative(nextProjectPath, testFile)], {
		cwd: nextProjectPath,
		env: {
			VERCEL_TEST_TEAM,
			VERCEL_TEST_TOKEN,
			NEXT_TEST_VERSION: url
		},
		stdio: "inherit"
	});
}

//#endregion
//#region src/commands/make.ts
const dependencies = {
	clean: [],
	install: [],
	build: ["clean", "install"],
	dev: [],
	types: [],
	native: [
		"clean",
		"install",
		"build"
	],
	default: [
		"clean",
		"install",
		"build",
		"dev"
	],
	all: ["native", "default"]
};
const create = (directory) => async (args, options) => {
	const nextProjectPath = await getConfig("next_project_path");
	let resolvedArgs = args;
	if (options?.filter) {
		resolvedArgs ??= [];
		resolvedArgs.unshift("--filter", ...options.filter);
	}
	return pnpm(resolvedArgs, {
		...options,
		stdio: "inherit",
		cwd: directory ? path.join(nextProjectPath, directory) : nextProjectPath
	});
};
const service = {
	project: create(),
	next: create("packages/next"),
	swc: create("packages/next-swc")
};
const commands = {
	types: async () => {},
	clean: async (options) => {
		if (!options.clean && !options.command.includes("clean")) return;
		await service.project(["clean"], options);
	},
	install: async (options) => {
		await service.project(["install"], options);
	},
	build: async (options) => {
		await service.project(["build"], options);
	},
	dev: async (options) => {
		await service.next(["dev"], options);
	},
	native: async (options) => {
		await service.swc(["build-native-release"], options);
	},
	default: null,
	all: null
};
/**
* Gathers the listed commands dependencies in required order of invocation.
* @param request
* @returns
*/
const gather = (request) => {
	const result = [];
	const helper = (command) => {
		for (const dep of dependencies[command]) if (!result.includes(dep)) helper(dep);
		if (!result.includes(command) && commands[command]) result.push(command);
	};
	request.forEach(helper);
	return result;
};
async function makeCommand(options) {
	const gathered = gather(options.command);
	for (const command of gathered) {
		const handler = commands[command];
		if (!handler) continue;
		await handler(options);
	}
}

//#endregion
//#region src/commands/next.ts
async function nextCommand(options) {
	const nextProjectPath = options.nextProjectDirectory;
	if (!existsSync(nextProjectPath)) throw new Error(`The next project directory ${nextProjectPath} does not exist.`);
	await next([options.command, nextProjectPath], { stdio: "inherit" });
}

//#endregion
//#region src/cli.ts
dotenv.config();
yargs(hideBin(process.argv)).command("test-deploy <test-file>", "performs a pack and test deploy of the specified test", {
	"no-app-dir": {
		type: "boolean",
		default: false
	},
	"skip-pack": {
		type: "boolean",
		default: false
	},
	"test-file": {
		type: "string",
		demand: true
	}
}, testDeployCommand).command("pack-next", "packs up and uploads the Next.js package to cloud storage", {
	json: {
		type: "boolean",
		default: false
	},
	serve: {
		type: "boolean",
		default: false
	},
	install: {
		type: "boolean",
		default: false
	}
}, packNextCommand).command("pack", "packs up and uploads the current package to cloud storage", {
	json: {
		type: "boolean",
		default: false
	},
	serve: {
		type: "boolean",
		default: false
	},
	progress: {
		type: "boolean",
		default: false
	},
	verbose: {
		type: "boolean",
		default: false
	}
}, packCommand).command("config <operation> [key] [value]", "update or get a config value", {
	operation: {
		choices: ["get", "set"],
		demand: true
	},
	key: { choices: Object.keys(schema) },
	value: { type: "string" }
}, configCommand).command("create-reproduction <name>", "creates a bare-bones reproduction project", {
	name: {
		type: "string",
		demand: true
	},
	"no-app-dir": {
		type: "boolean",
		default: false
	}
}, createReproductionCommand).command("make [command...]", "starts development with Next.js", {
	command: {
		choices: [
			"clean",
			"install",
			"build",
			"dev",
			"native",
			"default",
			"all"
		],
		default: ["default"],
		type: "array"
	},
	clean: {
		type: "boolean",
		default: false,
		description: "run `pnpm clean` before building"
	},
	filter: {
		type: "string",
		array: true,
		description: "filter the packages to run the command on"
	}
}, makeCommand).command("next <command> [next-project-directory]", "run commands using the development Next.js binary", {
	command: {
		type: "string",
		demand: true
	},
	"next-project-directory": {
		type: "string",
		normalize: true,
		default: process.cwd()
	}
}, nextCommand).command("debug <mode> <next-project-directory>", "debug a project with next", {
	mode: {
		choices: [
			"dev",
			"prod",
			"build",
			"start",
			"standalone",
			"export"
		],
		demand: true,
		description: "the task to run"
	},
	"next-project-directory": {
		type: "string",
		normalize: true,
		demand: true
	},
	rm: {
		type: "boolean",
		default: false
	},
	run: {
		type: "string",
		describe: "run a script/program while the server is running, signals will be sent to the child process",
		normalize: true
	}
}, debugCommand).demandCommand(1).strict().parse();

//#endregion